''' API for BalsamJob database (DAG) Manipulations
==================================================

The ``launcher.dag`` module provides a number of convenient environment
variables and functions that allow you to quickly write pre- and post-processing
scripts that interact with the BalsamJob database.

When pre- or post-processing steps for a job occur, the Launcher runs your
scripts in the job's working directory with some job-specific environment
variables. If you choose to write these scripts in Python with ``launcher.dag``,
then these variables are automatically loaded for you on module import.

Useful module-level attributes
-------------------------------

``current_job``
    The BalsamJob object which is currently being processed

``parents``
    Parent BalsamJob objects on which current_job depends

``children``
    Children BalsamJob objects that depend on current_job

``JOB_ID``
    Unique identifier and primary key of the BalsamJob

``TIMEOUT``
    Boolean flag indicating whether the current job has timed-out while running
    in the Launcher.  If True, Balsam has invoked the your script as a
    timeout-handler, and the script should take some clean-up or rescue acction
    such as spawning a new job.

``ERROR``
    Boolean flag indicating whether the current job's application returned a
    nonzero exit code to the Launcher.  If True, Balsam has invoked your
    script as an error-handler, and the script should take some clean-up or
    rescue action.

Usage example
--------------
A typical user's post-processing script might import and use the ``dag`` API as
follows::

     import balsam.launcher.dag as dag

     output = open('expected_output').read() # in job workdir

     if 'CONVERGED' not in output:
         # Kill subtree of this job
         for child in dag.children:
             dag.kill(child, recursive=True)

         # Create a child clone job with increased walltime and new input
         with open("input_rescue.dat", 'w') as fp:
             fp.write("# a new input file here")

         dag.spawn_child(clone=True,
             walltime_minutes=dag.current_job.walltime_minutes + 10, 
             input_files = 'input_rescue.dat')
'''
import json
import os
import uuid 
from collections import deque

from balsam import setup
setup()
from balsam.core.models import BalsamJob, history_line
from balsam.service.schedulers import JobEnv

__all__ = ['JOB_ID', 'TIMEOUT', 'ERROR', 
           'current_job', 'parents', 'children',
           'add_job', 'add_dependency', 'spawn_child',
           'kill']

current_job = None
parents = None
children = None
_envs = {k:v for k,v in os.environ.items() if k.find('BALSAM')>=0}

JOB_ID = _envs.get('BALSAM_JOB_ID', '')
TIMEOUT = _envs.get('BALSAM_JOB_TIMEOUT', False) == "TRUE"
ERROR = _envs.get('BALSAM_JOB_ERROR', False) == "TRUE"
LAUNCHER_NODES = JobEnv.num_workers
if LAUNCHER_NODES is None: 
    LAUNCHER_NODES = 1

if JOB_ID:
    JOB_ID = uuid.UUID(JOB_ID)
    try:
        current_job = BalsamJob.objects.get(pk=JOB_ID)
    except:
        raise RuntimeError(f"The environment specified current job: "
                           f"BALSAM_JOB_ID {JOB_ID}\n but this does not "
                           f"exist in DB! Was it deleted accidentally?"
                           "You may need to `unset BALSAM_JOB_ID`")
    else:
        parents = current_job.get_parents()
        children = current_job.get_children()

def add_job(
        name, workflow, application, 
        description='', args='',
        num_nodes=1, ranks_per_node=1, 
        cpu_affinity='depth', threads_per_rank=1,
        threads_per_core=1, 
        environ_vars={}, 
        data=None,
        save=True,
        **kwargs
    ):
    '''Add a new job to the BalsamJob DB
    
    Creates a new job and saves it to the database in CREATED state.
    The job is initialized with all blank/default values for its fields; these
    must be configured by the user or provided via ``kwargs``
    
    Args:
        - ``kwargs`` (*dict*): contains BalsamJob fields (keys) and their values to
          be set on BalsamJob instantiation.

    Returns:
        - ``job`` (*BalsamJob*): the newly-created BalsamJob instance

    Raises:
        - ``ValueError``: if an invalid field name is provided to *kwargs*
    '''
    job = BalsamJob()
    job.name             = name
    job.workflow         = workflow
    job.application      = application
    job.description      = description
    job.args             = args
    job.num_nodes        = num_nodes
    job.ranks_per_node   = ranks_per_node
    job.threads_per_rank = threads_per_rank
    job.threads_per_core = threads_per_core
    job.cpu_affinity   = cpu_affinity
    job.environ_vars   = environ_vars
    job.data = data if data else dict()
    job.get_application()

    for k,v in kwargs.items():
        setattr(job, k, v)

    if current_job:
        job.queued_launch = current_job.queued_launch
    if save:
        job.save()
    return job

def detect_circular(job, path=[]):
    '''Detect a circular dependency in DAG
    
    Args:
        - ``job`` (*BalsamJob*): node at which to start traversing the DAG
    
    Returns:
        - ``detected`` (*bool*): True if a circular dependency was detected
    ''' 
    if job.pk in path: return True
    path = path[:] + [job.pk]
    for parent in job.get_parents():
        if detect_circular(parent, path): return True
    return False

def breadth_first_iterator(roots, max_depth=None):
    try: 
        roots = iter(roots)
    except TypeError:
        assert isinstance(roots, BalsamJob)
        roots = [roots]

    queue = deque(roots)
    visited = []
    while queue:
        pass

def wf_from_template(source_wf, new_wf, **kwargs):
    source_jobs = BalsamJob.objects.filter(workflow=source_wf)
    if not source_jobs.exists():
        raise ValueError(f"No jobs exist matching workflow {source_wf}; \
                please provide an existing name for source_wf")
    if BalsamJob.objects.filter(workflow=new_wf).exists():
        raise ValueError(f"At least one job with workflow {new_wf} already exists;\
                please provide a unique name for new_wf")

    source_jobs = list(source_jobs)
    new_jobs = {}

    for job in source_jobs:
        kw = kwargs.get(job.name, {})
        kw.update(name=job.name, workflow=new_wf, queued_launch=None, parents="[]")
        new_job = clone(job, **kw)
        assert new_job.name not in new_jobs
        new_jobs[new_job.name] = new_job

    for j in new_jobs.values():
        j.state_history = history_line("CREATED", f"Copied from template workflow {source_wf}")
        j.save()

    for old_job in source_jobs:
        new_job = new_jobs[old_job.name]
        parent_names = old_job.get_parents().values_list('name', flat=True)
        for parent_name in parent_names:
            new_parent = new_jobs[parent_name]
            add_dependency(parent=new_parent, child=new_job)

    return BalsamJob.objects.filter(workflow=new_wf)

def add_dependency(parent,child):
    '''Create a dependency between two existing jobs
    
    Args:
        - ``parent`` (*BalsamJob*): The job which must reach state JOB_FINISHED
          before ``child`` begins processing
        - ``child`` (*BalsamJob*): The job that is dependent on ``parent`` for
          control- and/or data-flow.

    Raises:
        - ``RuntimeError``: if the attempted edge would create a circular
          dependency in the BalsamJob DAG.
    '''
    from django.db.models.query import QuerySet
    if isinstance(parent, str):
        parent = uuid.UUID(parent)
        parent = BalsamJob.objects.get(pk=parent)
    elif isinstance(parent, uuid.UUID):
        parent = BalsamJob.objects.get(pk=parent)
    elif isinstance(parent, QuerySet): 
        assert parent.count() == 1
        parent = parent.first()
    else:
        assert isinstance(parent, BalsamJob)

    if isinstance(child, str): 
        child = uuid.UUID(child)
        child = BalsamJob.objects.get(pk=child)
    elif isinstance(child, uuid.UUID):
        child = BalsamJob.objects.get(pk=child)
    elif isinstance(child, QuerySet): 
        assert child.count() == 1
        child = child.first()
    else:
        assert isinstance(child, BalsamJob)

    existing_parents = child.get_parents_by_id()
    new_parents = existing_parents.copy()
    parent_pk_str = str(parent.pk)
    if parent_pk_str in existing_parents:
        raise RuntimeError("Dependency already exists; cannot double-create")
    else:
        new_parents.append(parent_pk_str)
    child.set_parents(new_parents)
    if detect_circular(child):
        child.set_parents(existing_parents)
        raise RuntimeError("Detected circular dependency; not creating link")

def clone(job, **kwargs):
    assert isinstance(job, BalsamJob)
    new_job = BalsamJob()
    
    exclude_fields = '''_state objects source state tick user_workdir
    lock state_history job_id'''.split()
    fields = [f for f in job.__dict__ if f not in exclude_fields]

    for f in fields: 
        new_job.__dict__[f] = job.__dict__[f]
    assert new_job.pk != job.pk
    
    for k,v in kwargs.items():
        try: field = job._meta.get_field(k)
        except: raise ValueError(f"Invalid field name: {k}")
        else: new_job.__dict__[k] = v
    return new_job

def spawn_child(**kwargs):
    '''Add a new job that is dependent on the current job
    
    This function creates a new child job that will not start until the current
    job is finished processing. The job is added to the BalsamJob database in
    CREATED state.

    Args:
        - ``clone`` (*bool*): If *True*, all fields of the current BalsamJob are
          copied into the child job (except for primary key and working
          directory). Specific fields may then be overwritten via *kwargs*. 
          Defaults to *False*.
        - ``kwargs`` (*dict*) : Contains BalsamJob field names as keys and their
          desired values.

    Returns:
        - ``child`` (BalsamJob): returns the newly created BalsamJob

    Raises:
        - ``RuntimeError``: If no BalsamJob detected on module-load
        - ``ValueError``: if an invalid field name is passed into *kwargs* 
    '''
    if not isinstance(current_job, BalsamJob):
        raise RuntimeError("No current BalsamJob detected in environment")

    if 'workflow' not in kwargs:
        kwargs['workflow'] = current_job.workflow
    
    child = clone(current_job, **kwargs)
    child.queued_launch = current_job.queued_launch

    newparents = json.loads(current_job.parents)
    newparents.append(str(current_job.job_id))
    child.parents = json.dumps(newparents)
    child.state = "CREATED"
    child.state_history = history_line("CREATED", f"spawned by {current_job.cute_id}")
    child.save()
    return child

def kill(job, recursive=True):
    '''Kill a job or its entire subtree in the DAG

    Mark a job (and optionally all jobs that depend on it) by the state
    USER_KILLED, which will prevent any further processing.
    
    Args:
        - ``job`` (*BalsamJob*): the job (or subtree root) to kill
        - ``recursive`` (*bool*): if *True*, then traverse the DAG recursively
          to kill all jobs in the subtree. Defaults to *True*
    '''
    job.update_state('USER_KILLED')
    if recursive:
        for child in job.get_children():
            kill(child, recursive=True)

def get_database_paths(verbose=True):
    """
    Prints the paths for existing balsam databases
    """
    try:
        from balsam.django_config.db_index import refresh_db_index
        databasepaths = refresh_db_index()
    except:
        databasepaths = None
    if verbose:
        if len(databasepaths) > 0:
            print(f'Found {len(databasepaths)} balsam database location')
            for db in databasepaths:
                print(db)
        else:
            print('No balsam database found')
    return databasepaths

def get_active_database(verbose=True):
    """
    Gets the activate database set in environment variable BALSAM_DB_PATH
    Parameters:
    verbose: Boolean, (True): Prints verbose info (False): No print
    Returns
    -------
    str, path for the active database
    """
    db = os.environ.get("BALSAM_DB_PATH")
    if verbose: print(f'Active balsam database path: {db}')
    return db

def add_app(name, executable, description='', envscript='', preprocess='', postprocess='', checkexe=False):
    """
    Adds a new app to the balsam database.
    """
    from balsam.core.models import ApplicationDefinition as App
    import shutil
    
    if checkexe and not shutil.which(executable):        
        raise ValueError('No executable {} found in the PATH'.format(executable))

    newapp, created = App.objects.get_or_create(name=name)
    newapp.name        = name
    newapp.executable  = executable
    newapp.description = description
    newapp.envscript   = envscript
    newapp.preprocess  = preprocess
    newapp.postprocess = postprocess
    newapp.save()
    if created: print("Created new app")
    else: print("Updated existing app")
    return newapp

def get_apps(verbose=True):
    """
    Returns all apps as a list
    """
    try:
        from balsam.core.models import ApplicationDefinition as App
        apps = App.objects.all()
    except:
        apps = None
    return apps

def submit(project='datascience',queue='debug-flat-quad',nodes=1,wall_minutes=30,job_mode='mpi',wf_filter=''):
    """
    Submits a job to the queue with the given parameters.
    Parameters
    ----------
    project: str, name of the project to be charged
    queue: str, queue name, can be: 'default', 'debug-cache-quad', or 'debug-flat-quad'
    nodes: int, Number of nodes, can be any integer from 1 to 4096.
    wall_minutes: int, max wall time in minutes
    job_mode: str, Balsam job mode, can be 'mpi', 'serial'
    wf_filter: str, Selects Balsam jobs that matches the given workflow filter.
    """
    from balsam.service import service
    from balsam.core import models
    QueuedLaunch = models.QueuedLaunch
    mylaunch = QueuedLaunch()
    mylaunch.project = project
    mylaunch.queue = queue
    mylaunch.nodes = nodes
    mylaunch.wall_minutes = wall_minutes
    mylaunch.job_mode = job_mode
    mylaunch.wf_filter = wf_filter
    mylaunch.prescheduled_only=False
    mylaunch.save()
    service.submit_qlaunch(mylaunch, verbose=True)



import os
from getpass import getuser
from datetime import datetime

from django.conf import settings
from balsam.service.schedulers.exceptions import * 
from balsam.service.schedulers import Scheduler

import logging
logger = logging.getLogger(__name__)

def new_scheduler():
    return CobaltScheduler()

class CobaltScheduler(Scheduler.Scheduler):
    SCHEDULER_VARIABLES = {
        'current_scheduler_id' : 'COBALT_JOBID',
        'num_workers'  : 'COBALT_PARTSIZE',
        'workers_str'  : 'COBALT_PARTNAME',
        'workers_file' : 'COBALT_NODEFILE',
    }
    JOBSTATUS_VARIABLES = {
        'id' : 'JobID',
        'time_remaining' : 'TimeRemaining',
        'wall_time' : 'WallTime',
        'state' : 'State',
        'queue' : 'Queue',
        'nodes' : 'Nodes',
        'project' : 'Project',
        'command' : 'Command',
    }
    QSTAT_EXE = settings.SCHEDULER_STATUS_EXE

    def _make_submit_cmd(self, script_path):
        exe = settings.SCHEDULER_SUBMIT_EXE # qsub
        cwd = settings.SERVICE_PATH
        basename = os.path.basename(script_path)
        basename = os.path.splitext(basename)[0]
        return f"{exe} --cwd {cwd} -O {basename} {script_path}"

    def _parse_submit_output(self, submit_output):
        try: scheduler_id = int(submit_output)
        except ValueError: scheduler_id = int(submit_output.split()[-1])
        return scheduler_id

    def _make_status_cmd(self):
        fields = self.JOBSTATUS_VARIABLES.values()
        cmd = "QSTAT_HEADER=" + ':'.join(fields)
        cmd += f" {self.QSTAT_EXE} -u {getuser()}"
        return cmd

    def _parse_status_output(self, raw_output):
        status_dict = {}
        job_lines = raw_output.split('\n')[2:]
        for line in job_lines:
            job_stat = self._parse_job_line(line)
            if job_stat:
                id = int(job_stat['id'])
                status_dict[id] = job_stat
        return status_dict

    def _parse_job_line(self, line):
        fields = line.split()
        num_expected = len(self.JOBSTATUS_VARIABLES)
        if len(fields) != num_expected: return {}
        stat = {}
        for i, field_name in enumerate(self.JOBSTATUS_VARIABLES.keys()):
            stat[field_name] = fields[i]
            if 'time' in field_name:
                try:
                    t = datetime.strptime(fields[i], '%H:%M:%S')
                except:
                    pass
                else:
                    tsec = t.hour*3600 + t.minute*60 + t.second
                    stat[field_name+"_sec"] = tsec
                    tmin = t.hour*60 + t.minute
                    stat[field_name+"_min"] = tmin
        logger.debug(str(stat))
        return stat



import random
import numpy as np
import cython

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int any_occupied(int [:,:] a, int x1, int x2, int y1, int y2):
    cdef int ix, iy
    for ix in range(x1, x2+1):
        for iy in range(y1, y2+1):
            if a[ix, iy]: return 1
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
cdef int bisect_left(double [:] a, int n,  double x):
    cdef Py_ssize_t lo, hi, mid
    lo = 0
    hi = n
    while lo < hi:
        mid = (lo+hi)//2
        if a[mid] < x: lo = mid+1
        else: hi = mid
    return lo

class Rect:
    def __init__(self, xdim, ydim, id):
        self.xdim = xdim
        self.ydim = ydim
        self.id = id
    def __repr__(self):
        return f'<Rect{self.id}: ({self.xdim}, {self.ydim})>'

class BinPacker:
    def __init__(self, max_x, max_y):
        self.splits_x = np.array([0, max_x], dtype=np.float64)
        self.splits_y = np.array([0, max_y], dtype=np.float64)
        self.max_x = max_x
        self.max_y = max_y
        self.grid = np.zeros((1,1), dtype=np.int32)
        self.placed_rects = []

    def add_rect(self, int ix, int iy, double split_x, double split_y, int br_ix, int br_iy):
        if split_x not in self.splits_x:
            self.splits_x = np.insert(self.splits_x, br_ix+1, split_x)
            self.grid = np.insert(self.grid, br_ix, self.grid[br_ix,:], axis=0)
        if split_y not in self.splits_y:
            self.splits_y = np.insert(self.splits_y, br_iy+1, split_y)
            self.grid = np.insert(self.grid, br_iy, self.grid[:,br_iy], axis=1)
        self.grid[ix:br_ix+1, iy:br_iy+1] = 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def try_place(self, rect, first_col=False):
        fit = None
        cdef int [:,:] grid
        cdef int ix, iy, ixmax, iymax, br_ix, br_iy, i, j
        cdef double xdim, ydim, max_x, max_y
        cdef double ul_x, ul_y, br_x, br_y
        cdef double [:] splits_x
        cdef double [:] splits_y
        cdef int nsplits_x, nsplits_y

        splits_x = self.splits_x
        splits_y = self.splits_y
        nsplits_x = len(splits_x)
        nsplits_y = len(splits_y) 
        xdim = rect.xdim
        ydim = rect.ydim
        max_x = self.max_x
        max_y = self.max_y
        grid = self.grid
        ixmax, iymax = self.grid.shape
        if first_col: iymax = 1
        for iy in range(0, iymax):
            for ix in range(0, ixmax):
                if grid[ix, iy]: continue
                ul_x = splits_x[ix]
                ul_y = splits_y[iy]
                br_x = ul_x + xdim
                br_y = ul_y + ydim
                if br_x > max_x or br_y > max_y: continue
                br_ix = bisect_left(splits_x, nsplits_x, br_x) - 1 # using custom-C bisect is performance-critical
                br_iy = bisect_left(splits_y, nsplits_y, br_y) - 1
                if not any_occupied(grid, ix, br_ix, iy, br_iy):
                    self.add_rect(ix, iy, br_x, br_y, br_ix, br_iy)
                    ul = self.splits_x[ix], self.splits_y[iy]
                    self.placed_rects.append((ul, rect))
                    return

    def shrink_x_to_fit(self): 
        occ_x, occ_y = np.where(self.grid)
        x_bound = self.splits_x[occ_x.max()+1]
        self.max_x = x_bound
        self.splits_x[-1] = self.max_x
    
    def shrink_y_to_fit(self): 
        occ_x, occ_y = np.where(self.grid)
        y_bound = self.splits_y[occ_y.max()+1]
        self.max_y = y_bound
        self.splits_y[-1] = self.max_y

    def report(self, draw=False):
        for ul, rect in self.placed_rects:
            print(rect, 'at', ul)
        if draw:
            self.draw()

    def draw(self):
        from matplotlib import pyplot as plt
        import matplotlib.patches as patches
        import matplotlib.cm as cm
        fig, ax = plt.subplots(1)

        colors = cm.rainbow(np.linspace(0,1, len(self.placed_rects)))
        title = f'Fit {len(self.placed_rects)} rects in bbox {self.max_x}x{self.max_y}'
        for c, (ul, rect) in zip(colors, self.placed_rects):
            ll = ul[0] + rect.xdim, ul[1]
            x, y = ll[1], self.max_x - ll[0]
            width, height = rect.ydim, rect.xdim
            r = patches.Rectangle((x,y), width, height, lw=1,
                                  edgecolor='k',alpha=0.5, facecolor=c)
            ax.add_patch(r)
            #ax.text(x+width/2, y+height/2, f'{rect.id}', fontsize=6)
        r = patches.Rectangle((0,0), self.max_y, self.max_x, lw=2, edgecolor='r', facecolor='none')
        ax.add_patch(r)
        ax.set_ylim(0, self.max_x)
        ax.set_xlim(0, self.max_y)
        ax.set_title(title)
        plt.show()

def main_draw(*, num_rect=500, xmax=2048, ymax=1440, rect_xrange=(16,128), rect_yrange=(30,100), draw_interval=100):
    packer = BinPacker(xmax, ymax)
    rects = [Rect(random.randint(*rect_xrange), 
                  random.randint(*rect_yrange), 
                  i) 
             for i in range(num_rect)]

    for tot, r in enumerate(sorted(rects, key = lambda r: r.xdim, reverse=True), 1): 
        packer.try_place(r)
        if tot % draw_interval == 0: 
            packer.report(draw=True)



from collections import namedtuple, defaultdict
import time
from balsam import setup
setup()
from balsam.core.models import BalsamJob
from balsam.core.models import END_STATES
from balsam.launcher import dag

class TaskFailed(Exception): pass
WaitResult = namedtuple('WaitResult', ['active', 'done', 'failed', 'cancelled'])

def _timer(timeout):
    if timeout is None: 
        return lambda : True
    else:
        timeout = max(float(timeout), 0.01)
        start = time.time()
        return lambda : (time.time()-start) < timeout

def _to_state(state):
    if state not in END_STATES:
        return 'active'
    elif state == 'JOB_FINISHED':
        return 'done'
    elif state == 'FAILED':
        return 'failed'
    else:
        assert state == 'USER_KILLED'
        return 'cancelled'

def wait(futures, timeout=None, return_when='ANY_COMPLETED', poll_period=1.0):
    assert return_when.strip() in ['ANY_COMPLETED', 'ALL_COMPLETED']
    waitall = bool(return_when.strip() == 'ALL_COMPLETED')

    time_isLeft = _timer(timeout)
    futures = {future._job.pk:future for future in futures}
    finished_pks = []

    if waitall: can_exit = lambda : len(finished_pks) == len(futures)
    else: can_exit = lambda : len(finished_pks) > 0

    while time_isLeft():
        pks = [pk for pk in futures if pk not in finished_pks]
        jobs = BalsamJob.objects.filter(pk__in=pks)
        for j in jobs:
            futures[j.pk]._job = j
            futures[j.pk]._state = _to_state(j.state)
            if j.state in END_STATES: finished_pks.append(j.pk)
        if can_exit(): break
        else: time.sleep(poll_period)

    if not can_exit():
        raise TimeoutError(f'{timeout} sec timeout expired while '
        f'waiting on {len(futures)} tasks until {return_when}')

    results = defaultdict(list)
    for f in futures.values(): results[f._state].append(f)
    return WaitResult(
        active=results['active'],
        done=results['done'],
        failed=results['failed'],
        cancelled=results['cancelled']
    )

class FutureTask:
    def __init__(self, balsam_job, done_callback, fail_callback=None, *args, **kwargs):
        '''Create a FutureTask from a BalsamJob object and callback function.

        Args
            balsam_job: BalsamJob object to be wrapped and monitored
            done_callback: a function of one positional argument (BalsamJob) and
                optional *args, **kwargs. done_callback(balsam_job, *args,
                **kwargs) is invoked and its return value is returned upon calling
                the result() method of a FutureTask.
            fail_callback (optional): a function with the same signature as done_callback,
                which is invoked upon calling result() if the task fails. 
            *args, **kwargs: optional args and keyword args to the done and fail
            callbacks.
        '''
        assert isinstance(balsam_job, BalsamJob)
        self._job = balsam_job
        self._on_done = done_callback
        self._on_fail = fail_callback
        assert callable(self._on_done)

        self._result = None
        self._state = 'active'
        self._executed_callback = False
        self._args = args
        self._kwargs = kwargs

    def __repr__(self):
        return f'<balsam.launcher.futures.FutureTask {self._job.cute_id}: {self._state}>' 

    def _poll(self):
        self._job = BalsamJob.objects.get(pk=self._job.pk)
        self._state = _to_state(self._job.state)

    def _callback(self, success):
        callback = self._on_done if success else self._on_fail
        if not success and not callable(callback):
            raise TaskFailed(f'Task {self._job.cute_id} failed; no fail_callback was set to handle this.')
        if not self._executed_callback:
            self._result = callback(self._job, *self._args, **self._kwargs)
            self._executed_callback = True

    @property
    def done(self):
        if self._state == 'active': self._poll()
        return self._state == 'done'

    @property
    def failed(self):
        if self._state == 'active': self._poll()
        return self._state == 'failed'
    
    @property
    def cancelled(self):
        if self._state == 'active': self._poll()
        return self._state == 'cancelled'

    @property
    def active(self):
        if self._state == 'active': self._poll()
        return self._state == 'active'

    def result(self, timeout=None):
        if self._executed_callback: return self._result

        time_isLeft = _timer(timeout)

        while time_isLeft():
            if self.active: time.sleep(1)
            else: break

        if self.active:
            raise TimeoutError(f"Task {self.cute_id} not finished after {timeout} seconds")
        elif self.cancelled:
            return None
        elif self.done:
            self._callback(True)
            return self._result
        else:
            assert self.failed
            self._callback(False)
            return self._result

    def cancel(self):
        dag.kill(self._job)
        self._state = 'cancelled'



from django.conf.urls import url,include
from balsam.core import views as balsam_views
from balsam.core import urls as balsam_urls

urlpatterns = [
    url(r'^$', balsam_views.home_page, name="home"),
    url(r'^balsam/', include(balsam_urls)),
]



from getpass import getuser
import os
from socket import gethostname

from django.shortcuts import render, redirect
from balsam.core.models import BalsamJob

def home_page(requests):
    return redirect('/balsam/tasks')

def info_str():
    user = getuser()
    host = gethostname()
    db_name = os.path.basename(os.environ.get('BALSAM_DB_PATH'))
    return f"<p>{user}@{host}</p>Balsam DB: {db_name}"

def list_tasks(request):
    info = info_str()
    return render(request, 'tasks.html', {'db_info':info})

def list_apps(request):
    info = info_str()
    return render(request, 'apps.html', {'db_info':info})



from mpi4py import MPI
import balsam.launcher.dag as dag

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

job_name = f"hello{rank}"
dag.add_job(name=job_name, workflow="test", application="hello", num_nodes=1,
            ranks_per_node=1)
print(f"Rank {rank} added job: success")



import json
import os
import subprocess
import socket
from getpass import getuser

ADDRESS_FNAME = 'server-info'


class ServerInfo:
    def __init__(self, balsam_db_path):
        self.balsam_db_path = os.path.abspath(os.path.expanduser(balsam_db_path))
        self.pg_db_path = os.path.join(self.balsam_db_path, 'balsamdb')
        self.path = os.path.join(self.balsam_db_path, ADDRESS_FNAME)
        self.data = {}
        if not os.path.exists(self.path):
            self._is_owner = True
            self.update(self.data)
        else:
            self._is_owner = os.stat(self.path).st_uid == os.getuid()
            self.refresh()

    def reset_server_address(self):
        self._ownership_check()
        info = self._conn_info()
        self.update(info)
        self._update_postgres_config()

    def update(self, update_dict):
        if os.path.exists(self.path): self.refresh()
        self.data.update(update_dict)
        with open(self.path, 'w') as fp:
            fp.write(json.dumps(self.data))


    def django_db_config(self):
        ENGINE = 'django.db.backends.postgresql_psycopg2'
        NAME = 'balsam'
        OPTIONS = {'connect_timeout' : 30}

        user = os.environ.get('BALSAM_USER', getuser())
        password = self.get('password', '')
        host = self.get('host', '')
        port = self.get('port', '')


        db = dict(ENGINE=ENGINE, NAME=NAME,
                  OPTIONS=OPTIONS, USER=user, PASSWORD=password,
                  HOST=host, PORT=port, CONN_MAX_AGE=60)

        DATABASES = {'default':db}
        return DATABASES

    def get(self, key, default=None):
        if key in self.data:
            return self.data[key]
        else:
            return default

    def refresh(self):
        with open(self.path, 'r') as fp:
            file_data = json.loads(fp.read())
        self.data.update(file_data)

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.update({key:value})
    
    def _ownership_check(self):
        if not self._is_owner:
            raise PermissionError(f"User {getuser()} does not own the server-info file for this BalsamDB;"
                                  " will not proceed")

    def _free_port(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.bind(('', 0))
        port = int(sock.getsockname()[1])
        sock.close()
        return port

    def _conn_info(self):
        hostname = socket.gethostname()
        port = self._free_port()
        info = dict(host=hostname, port=port)
        return info

    def _update_postgres_config(self):
        self._ownership_check()
        conf_path = os.path.join(self.pg_db_path, 'postgresql.conf')
        config = open(conf_path).read()

        with open(f"{conf_path}.new", 'w') as fp:
            for line in config.split('\n'):
                if line.startswith('port'):
                    port_line = f"port={self['port']} # auto-set by balsam db\n"
                    fp.write(port_line)
                else:
                    fp.write(line + "\n")
        os.rename(f"{conf_path}.new", conf_path)



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import glob
import balsam.launcher.dag as dag

square_files = glob.glob("square*.dat*")
job = dag.current_job

job.application_args=" ".join(square_files)
print(f"Have {len(square_files)} squares to sum")
job.save(update_fields=['application_args'])



'''Worker: Abstraction for the compute unit running a job.
Theta: 1 worker = 1 node
BG/Q: 1 worker = 1 subblock
Default: 1 worker = local host machine

Workers contain any identifying information needed to assign jobs to specific
workers (e.g. via "mpirun") and the WorkerGroup keeps track of all busy and idle
workers available in the current launcher instance'''

import logging
logger = logging.getLogger(__name__)

from balsam.service.schedulers import JobEnv
from balsam.launcher import mpi_commands

class Worker:
    def __init__(self, id, *, shape=None, block=None, corner=None,
                 num_nodes=None, host_type=None):
        self.id = id
        self.shape = shape
        self.block = block
        self.corner = corner
        self.num_nodes = num_nodes
        self.host_type = host_type
        self.idle = True
    def __repr__(self):
        return f"worker{self.id}"

class WorkerGroup:
    '''Collection of Workers, constructed by passing in a specific host_type
    
    The host name and local batch scheduler's environment variables are used to
    identify the available compute resources and partition the resource into
    Workers.'''
    def __init__(self):
        '''Initialize WorkerGroup
        
        Args:
            - ``host_type``: one of THETA, BGQ, COOLEY, DEFAULT
            - ``workers_str``: system-specific string identifying compute
              resources
            - ``workers_file``: system-specific file identifying compute
              resources
        '''
        self.host_type = JobEnv.host_type
        self.workers_str = JobEnv.workers_str
        self.workers_file = JobEnv.workers_file
        self.workers = []
        self.setup = getattr(self, f"setup_{self.host_type}")
        self.setup()
        self.mpi_cmd = mpi_commands.MPIcmd()

        logger.info(f"Built {len(self.workers)} {self.host_type} workers")
        for worker in self.workers:
            worker.mpi_cmd = self.mpi_cmd
            logger.debug(f"ID {worker.id} NODES {worker.num_nodes}")

    def __iter__(self):
        return iter(self.workers)

    def __len__(self):
        return len(self.workers)

    def idle_workers(self):
        return [w for w in self.workers if w.idle]

    def request(self, num_nodes):
        idle_workers = self.idle_workers()
        assigned = []
        nodes_assigned = 0
        while nodes_assigned < num_nodes:
            if not idle_workers: break
            worker = idle_workers.pop()
            assigned.append(worker)
            nodes_assigned += worker.num_nodes
        if nodes_assigned < num_nodes:
            return []
        else:
            for worker in assigned: worker.idle = False
            return assigned

    def __getitem__(self, i):
        return self.workers[i]

    def setup_THETA(self):
        # workers_str is string like: 1001-1005,1030,1034-1200
        node_ids = []
        if not self.workers_str:
            raise ValueError("Theta WorkerGroup needs workers_str to setup")
            
        ranges = self.workers_str.split(',')
        for node_range in ranges:
            lo, *hi = node_range.split('-')
            lo = int(lo)
            if hi:
                hi = int(hi[0])
                node_ids.extend(list(range(lo, hi+1)))
            else:
                node_ids.append(lo)
        for id in node_ids:
            self.workers.append(Worker(id, host_type='THETA', num_nodes=1))

    def setup_BGQ(self):
        # Boot blocks
        # Get (block, corner, shape) args for each sub-block
        # For each worker, set num_nodes
        pass

    def setup_COOLEY(self):
        node_ids = []
        if not self.workers_file:
            raise ValueError("Cooley WorkerGroup needs workers_file to setup")
        
        data = open(self.workers_file).read()
        splitter = ',' if ',' in data else None
        node_ids = data.split(splitter)
        self.workers_str = " ".join(node_ids)
        
        for id in node_ids:
            self.workers.append(Worker(id, host_type='COOLEY', num_nodes=1))

    def setup_DEFAULT(self):
        w = Worker(1, host_type='DEFAULT', num_nodes=1)
        self.workers.append(w)



from django.conf.urls import url
from balsam.core import views, api_views

urlpatterns = [
    url(r'^$', views.home_page, name="home"),
    url(r'^tasks/$', views.list_tasks, name="tasks"),
    url(r'^apps/', views.list_apps, name="apps"),
    url(r'^api/tasks_list', api_views.list_tasks, name="api_tasks"),
    url(r'^api/tasks/([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})/$', 
        api_views.task_detail, name="api_task_detail"),
]



class BalsamLauncherError(Exception): pass

class BalsamTransitionError(Exception): pass
class TransitionNotFoundError(BalsamTransitionError, ValueError): pass

class MPIEnsembleError(Exception): pass



import random
import numpy as np
import bisect
import time
import cProfile

class Rect:

    CURRENT_ID = 0

    def __init__(self, xdim, ydim, id=None):
        self.xdim = xdim
        self.ydim = ydim
        if id is not None: 
            self.id = id
        else:
            self.id = Rect.CURRENT_ID
            Rect.CURRENT_ID += 1

    def __repr__(self):
        return f'<Rect{self.id}: ({self.xdim}, {self.ydim})>'

    @classmethod
    def rand_rect(cls, width_range, height_range):
        w = random.randint(*width_range)
        h = random.randint(*height_range)
        rect = cls(w, h, cls.CURRENT_ID)
        cls.CURRENT_ID += 1
        return rect

class BinPacker:
    def __init__(self, max_x, max_y):
        self.splits_x = [0, max_x]
        self.splits_y = [0, max_y]
        self.max_x = max_x
        self.max_y = max_y
        self.grid = np.zeros((1,1), dtype=np.bool)
        self.placed_rects = []

    def empty_iter(self): 
        return sorted(np.argwhere(self.grid == False), key= lambda x: (x[1],x[0]))

    def check_fit(self, rect, ix, iy):
        ul = self.splits_x[ix], self.splits_y[iy]
        br = ul[0] + rect.xdim, ul[1] + rect.ydim
        if br[0] > self.max_x or br[1] > self.max_y: return False

        br_ix = bisect.bisect_left(self.splits_x, br[0]) - 1
        br_iy = bisect.bisect_left(self.splits_y, br[1]) - 1
        
        if np.any(self.grid[ix:br_ix+1, iy:br_iy+1]): return False
        return br, (br_ix, br_iy)

    def add_rect(self, ix, iy, split_x, split_y, br_ix, br_iy):
        if split_x not in self.splits_x:
            self.splits_x.insert(br_ix+1, split_x)
            self.grid = np.insert(self.grid, br_ix, self.grid[br_ix,:], axis=0)
        if split_y not in self.splits_y:
            self.splits_y.insert(br_iy+1, split_y)
            self.grid = np.insert(self.grid, br_iy, self.grid[:,br_iy], axis=1)
        self.grid[ix:br_ix+1, iy:br_iy+1] = True

    def try_place(self, rect, first_col=False):
        fit = None
        for ix, iy in self.empty_iter():
            if first_col and iy > 0: break
            fit = self.check_fit(rect, ix, iy)
            if fit: break
        if fit:
            (split_x, split_y), (br_ix, br_iy) = fit
            self.add_rect(ix, iy, split_x, split_y, br_ix, br_iy)
            ul = self.splits_x[ix], self.splits_y[iy]
            self.placed_rects.append((ul, rect))
        else:
            #print(rect, 'did not fit')
            return

    def shrink_x_to_fit(self): 
        occ_x, occ_y = np.where(self.grid)
        x_bound = self.splits_x[occ_x.max()+1]
        self.max_x = x_bound
        self.splits_x[-1] = self.max_x
    
    def shrink_y_to_fit(self): 
        occ_x, occ_y = np.where(self.grid)
        y_bound = self.splits_y[occ_y.max()+1]
        self.max_y = y_bound
        self.splits_y[-1] = self.max_y

    def report(self, draw=False):
        for ul, rect in self.placed_rects:
            print(rect, 'at', ul)
        if draw:
            self.draw()

    def draw(self):
        from matplotlib import pyplot as plt
        import matplotlib.patches as patches
        import matplotlib.cm as cm
        fig, ax = plt.subplots(1)

        colors = cm.rainbow(np.linspace(0,1, len(self.placed_rects)))
        title = f'Fit {len(self.placed_rects)} rects in bbox {self.max_x}x{self.max_y}'
        for c, (ul, rect) in zip(colors, self.placed_rects):
            ll = ul[0] + rect.xdim, ul[1]
            x, y = ll[1], self.max_x - ll[0]
            width, height = rect.ydim, rect.xdim
            r = patches.Rectangle((x,y), width, height, lw=1,
                                  edgecolor='k',alpha=0.5, facecolor=c)
            ax.add_patch(r)
            #ax.text(x+width/2, y+height/2, f'{rect.id}', fontsize=6)
        r = patches.Rectangle((0,0), self.max_y, self.max_x, lw=2, edgecolor='r', facecolor='none')
        ax.add_patch(r)
        ax.set_ylim(0, self.max_x)
        ax.set_xlim(0, self.max_y)
        ax.set_title(title)
        plt.show()

def main_draw():
    NUM_RECT = 5000
    XMAX, YMAX = 2048, 1440
    RECT_XRANGE = (2, 128)
    RECT_YRANGE = (30, 180)
    DRAW_INTERVAL = 200
    packer = BinPacker(XMAX, YMAX)
    rects = [Rect.rand_rect(RECT_XRANGE, RECT_YRANGE) for i in range(NUM_RECT)]

    start = time.time()
    try:
        for tot, r in enumerate(sorted(rects, key = lambda r: r.xdim, reverse=True), 1): 
            packer.try_place(r)
            if tot % DRAW_INTERVAL == 0: packer.report(draw=True)
            print(tot)
    except KeyboardInterrupt:
        elapsed = time.time() - start
        print(f'packed {tot} in {elapsed} seconds')
        packer.report(draw=True)

def main_prof():
    NUM_RECT = 600
    XMAX, YMAX = 2048, 1440
    RECT_XRANGE = (2, 128)
    RECT_YRANGE = (30, 180)
    packer = BinPacker(XMAX, YMAX)
    rects = [Rect.rand_rect(RECT_XRANGE, RECT_YRANGE) for i in range(NUM_RECT)]
    for r in sorted(rects, key = lambda r: r.xdim, reverse=True): 
        packer.try_place(r)

def pack():
    packer = BinPacker(10, 10)
    rects = [Rect(2+random.randint(-1,3), 2+random.randint(-2,4)) for i in range(4)]
    for r in sorted(rects, key = lambda r: r.xdim, reverse=True):
        packer.try_place(r, first_col=True)
    packer.report(draw=True)
    packer.shrink_x_to_fit()
    packer.report(draw=True)
    packer.shrink_y_to_fit()
    packer.report(draw=True)

cProfile.run('main_prof()', sort='cumtime')
#main_draw()



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import balsam.launcher.dag as dag
import glob

print("hello from square_pre")
print(f"jobid: {dag.current_job.pk}")
if not dag.current_job.application_args:
    print("no input file set for this job. searching workdir...")
    infile = glob.glob("side*.dat*")[0]
    dag.current_job.application_args += infile
    dag.current_job.save()
    print("set square.py input to", infile)



'''Main Launcher entry point

The ``main()`` function contains the Launcher service loop, in which:
    1. Transitions are checked for completion and jobs states are updated
    2. Dependencies of waiting jobs are checked
    3. New transitions are added to the transitions queue
    4. The RunnerGroup is checked for jobs that have stopped execution
    5. A new Runner is created according to logic in create_next_runner

The ``sig_handler()`` handler is invoked either when the time limit is exceeded or
if the program receives a SIGTERM or SIGINT. This takes the necessary cleanup
actions and is guaranteed to execute only once through the EXIT_FLAG global
flag.
'''
import argparse
from collections import defaultdict
from importlib.util import find_spec
import logging
from math import floor
import os
import sys
import signal
import subprocess
import shlex
import time

from django import db
from balsam import config_logging, settings, setup
from balsam.core import transitions
from balsam.launcher import worker
from balsam.launcher.util import remaining_time_minutes, delay_generator, get_tail
from balsam.launcher.exceptions import *
from balsam.scripts.cli import config_launcher_subparser
from balsam.core import models

logger = logging.getLogger('balsam.launcher.launcher')
BalsamJob = models.BalsamJob
EXIT_FLAG = False

def sig_handler(signum, stack):
    global EXIT_FLAG
    EXIT_FLAG = True

class MPIRun:
    RUN_DELAY = 0.10 # 1000 jobs / 100 sec

    def __init__(self, job, workers):
        self.job = job
        self.workers = workers
        for w in self.workers: w.idle = False

        envs = job.get_envs() # dict
        app_cmd = job.app_cmd
        nranks = job.num_ranks
        affinity = job.cpu_affinity
        rpn = job.ranks_per_node
        tpr = job.threads_per_rank
        tpc = job.threads_per_core

        mpi_cmd = workers[0].mpi_cmd
        mpi_str = mpi_cmd(workers, app_cmd=app_cmd, envs={},
                               num_ranks=nranks, ranks_per_node=rpn,
                               cpu_affinity=affinity, threads_per_rank=tpr, threads_per_core=tpc)
        basename = job.name
        outname = os.path.join(job.working_directory, f"{basename}.out")
        self.outfile = open(outname, 'w+b')
        envscript = job.envscript
        if envscript:
            args = ' '.join(['source', envscript, '&&', mpi_str])
            shell = True
        else:
            args = shlex.split(mpi_str)
            shell = False
        logger.info(f"{job.cute_id} Popen (shell={shell}):\n{args}\n on workers: {workers}")
        self.process = subprocess.Popen(
                args=args,
                cwd=job.working_directory,
                stdout=self.outfile,
                stderr=subprocess.STDOUT,
                shell=shell,
                env=envs,
                )
        self.current_state = 'RUNNING'
        self.err_msg = None
        time.sleep(self.RUN_DELAY)

    def free_workers(self):
        for worker in self.workers:
            worker.idle = True

class MPILauncher:
    MAX_CONCURRENT_RUNS = settings.MAX_CONCURRENT_MPIRUNS

    def __init__(self, wf_name, time_limit_minutes, gpus_per_node):
        self.jobsource = BalsamJob.source
        self.jobsource.workflow = wf_name
        if wf_name:
            logger.info(f'Filtering jobs with workflow matching {wf_name}')
        else:
            logger.info('No workflow filter')

        self.jobsource.clear_stale_locks()
        self.jobsource.start_tick()
        self.worker_group = worker.WorkerGroup()
        self.total_nodes = sum(w.num_nodes for w in self.worker_group)
        os.environ['BALSAM_LAUNCHER_NODES'] = str(self.total_nodes)
        os.environ['BALSAM_JOB_MODE'] = "mpi"

        self.timer = remaining_time_minutes(time_limit_minutes)
        self.delayer = delay_generator()
        self.last_report = 0
        self.exit_counter = 0
        self.mpi_runs = []
        self.jobsource.check_qLaunch()
        if self.jobsource.qLaunch is not None:
            sched_id = self.jobsource.qLaunch.scheduler_id
            self.RUN_MESSAGE = f'Batch Scheduler ID: {sched_id}'
        else:
            self.RUN_MESSAGE = 'Not scheduled by service'

    def time_step(self):
        '''Pretty log of remaining time'''
        global EXIT_FLAG

        next(self.delayer)
        try:
            minutes_left = next(self.timer)
        except StopIteration:
            EXIT_FLAG = True
            return

        if minutes_left > 1e12:
            return
        whole_minutes = floor(minutes_left)
        whole_seconds = round((minutes_left - whole_minutes)*60)
        time_str = f"{whole_minutes:02d} min : {whole_seconds:02d} sec remaining"
        logger.info(time_str)

    def check_exit(self):
        global EXIT_FLAG
        try: 
            remaining_minutes = next(self.timer)
        except StopIteration:
            EXIT_FLAG = True
            logger.info("Out of time; preparing to exit")
            return
        if remaining_minutes <= 0:
            EXIT_FLAG = True
            logger.info("Out of time; preparing to exit")
            return
        if self.is_active: 
            logger.debug("Some runs are still active; will not quit")
            return
        processable = BalsamJob.objects.filter(state__in=models.PROCESSABLE_STATES)
        if self.jobsource.workflow:
            processable = processable.filter(workflow__contains=self.jobsource.workflow)
        if processable.count() > 0:
            logger.debug("Some BalsamJobs are still transitionable; will not quit")
            return
        if self.get_runnable().count() > 0:
            self.exit_counter = 0
            return
        else:
            self.exit_counter += 1
            logger.info(f"Nothing to do (exit counter {self.exit_counter}/10)")
        if self.exit_counter == 10:
            EXIT_FLAG = True

    def check_state(self, run):
        retcode = run.process.poll()
        if retcode is None:
            run.current_state = 'RUNNING'
        elif retcode == 0:
            logger.info(f"MPIRun {run.job.cute_id} done")
            run.current_state = 'RUN_DONE'
            run.outfile.close()
            run.free_workers()
        else:
            run.process.communicate()
            run.outfile.close()
            tail = get_tail(run.outfile.name)
            run.current_state = 'RUN_ERROR'
            run.err_msg = tail
            logger.info(f"MPIRun {run.job.cute_id} error code {retcode}:\n{tail}")
            run.free_workers()
        return run.current_state

    def timeout_kill(self, runs, timeout=10):
        for run in runs: run.process.terminate() #SIGTERM
        start = time.time()
        for run in runs:
            try: 
                run.process.wait(timeout=timeout)
            except: 
                break
            if time.time() - start > timeout: break
        for run in runs: 
            run.process.kill()
            run.free_workers()

    def update(self, timeout=False):
        by_states = defaultdict(list)
        for run in self.mpi_runs:
            state = self.check_state(run)
            by_states[state].append(run)

        done_pks = [r.job.pk for r in by_states['RUN_DONE']]
        BalsamJob.batch_update_state(done_pks, 'RUN_DONE')
        self.jobsource.release(done_pks)
        
        error_pks = [r.job.pk for r in by_states['RUN_ERROR']]
        with db.transaction.atomic():
            models.safe_select(BalsamJob.objects.filter(pk__in=error_pks))
            for run in by_states['RUN_ERROR']:
                run.job.refresh_from_db()
                run.job.update_state('RUN_ERROR', run.err_msg)
        self.jobsource.release(error_pks)
        
        active_pks = [r.job.pk for r in by_states['RUNNING']]
        if timeout: 
            self.timeout_kill(by_states['RUNNING'])
            BalsamJob.batch_update_state(active_pks, 'RUN_TIMEOUT')
            self.jobsource.release(active_pks)
        else:
            killquery = self.jobsource.filter(job_id__in=active_pks, state='USER_KILLED')
            kill_pks  = killquery.values_list('job_id', flat=True)
            to_kill = [run for run in by_states['RUNNING'] if run.job.pk in kill_pks]
            self.timeout_kill(to_kill)
            self.jobsource.release(kill_pks)
            for run in to_kill: by_states['RUNNING'].remove(run)

        if timeout:
            self.mpi_runs = []
        else:
            self.mpi_runs = by_states['RUNNING']
        
    def get_runnable(self):
        '''queryset: jobs that can finish on idle workers (disregarding time limits)'''
        manager = self.jobsource
        num_idle = len(self.worker_group.idle_workers())
        if num_idle == 0:
            logger.debug(f'No idle worker nodes to run jobs')
            return manager.none()
        else:
            logger.debug(f'{num_idle} idle worker nodes')

        return manager.get_runnable(
            max_nodes=num_idle,
            order_by=('-num_nodes', '-wall_time_minutes')
        )

    def report_constrained(self):
        now = time.time()
        elapsed = now - self.last_report
        if elapsed < 10:
            return
        else:
            self.last_report = now
        num_idle = len(self.worker_group.idle_workers())
        logger.info(f'{num_idle} idle worker nodes')
        all_runnable = BalsamJob.objects.filter(state__in=models.RUNNABLE_STATES)
        unlocked = all_runnable.filter(lock='')
        logger.info('No runnable jobs')
        logger.info(f'{all_runnable.count()} runnable jobs across entire Balsam DB')
        logger.info(f'{unlocked.count()} of these are unlocked')
        if self.jobsource.workflow:
            unlocked = unlocked.filter(workflow__contains=self.jobsource.workflow)
            logger.info(f'{unlocked.count()} of these match the current workflow filter')
        too_large = unlocked.filter(num_nodes__gt=num_idle).count()
        if too_large > 0:
            logger.info(f'{too_large} of these could run now; but require more than {num_idle} nodes.')

    def launch(self):
        num_idle = len(self.worker_group.idle_workers())
        num_active = len(self.mpi_runs)
        max_acquire = min(num_idle, self.MAX_CONCURRENT_RUNS - num_active)
        max_acquire = max(max_acquire, 0)

        if num_active >= self.MAX_CONCURRENT_RUNS:
            logger.info(f'Reached MAX_CONCURRENT_MPIRUNS limit')
            return

        runnable = self.get_runnable()
        num_runnable = runnable.count()
        if num_runnable > 0:
            logger.debug(f"{num_runnable} runnable jobs")
        else:
            self.report_constrained()
            return

        # pre-assign jobs to nodes (descending order of node count)
        cache = list(runnable[:max_acquire])
        pre_assignments = []
        idx = 0
        while idx < len(cache):
            job = cache[idx]
            workers = self.worker_group.request(job.num_nodes)
            if workers:
                pre_assignments.append((job, workers))
                idx += 1
            else:
                num_idle = sum(w.num_nodes for w in self.worker_group.idle_workers())
                assert job.num_nodes > num_idle
                idx = next( (i for i,job in enumerate(cache[idx:], idx) if
                           job.num_nodes <= num_idle), len(cache))

        # acquire lock on jobs
        to_acquire = [job.pk for (job,workers) in pre_assignments]
        acquired_pks = self.jobsource.acquire(to_acquire)
        logger.debug(f'Acquired lock on {len(acquired_pks)} out of {len(pre_assignments)} jobs marked for running')

        # dispatch runners; release workers that did not acquire job
        for (job, workers) in pre_assignments:
            if job.pk in acquired_pks:
                run = MPIRun(job, workers)
                self.mpi_runs.append(run)
            else:
                for w in workers: w.idle = True
        BalsamJob.batch_update_state(acquired_pks, 'RUNNING', self.RUN_MESSAGE)

    def run(self):
        '''Main Launcher service loop'''
        global EXIT_FLAG
        try:
            while not EXIT_FLAG:
                self.time_step()
                self.launch()
                self.update()
                self.check_exit()
        except:
            raise
        finally:
            logger.debug('EXIT: breaking launcher service loop')
            self.update(timeout=True)
            assert not self.is_active
            logger.info('Exit: All MPI runs terminated')
            self.jobsource.release_all_owned()
            logger.info('Exit: Launcher Released all BalsamJob locks')

    @property
    def is_active(self):
        return len(self.mpi_runs) > 0


class SerialLauncher:
    MPI_ENSEMBLE_EXE = find_spec("balsam.launcher.mpi_ensemble").origin

    def __init__(self, wf_name=None, time_limit_minutes=60, gpus_per_node=None):
        self.wf_name = wf_name
        self.gpus_per_node = gpus_per_node

        timer = remaining_time_minutes(time_limit_minutes)
        minutes_left = max(0.1, next(timer) - 1)
        self.worker_group = worker.WorkerGroup()
        self.total_nodes = sum(w.num_nodes for w in self.worker_group)
        os.environ['BALSAM_LAUNCHER_NODES'] = str(self.total_nodes)
        os.environ['BALSAM_JOB_MODE'] = "serial"

        self.app_cmd = f"{sys.executable} {self.MPI_ENSEMBLE_EXE}"
        self.app_cmd += f" --time-limit-min={minutes_left}"
        if self.wf_name: self.app_cmd += f" --wf-name={self.wf_name}"
        if self.gpus_per_node: self.app_cmd += f" --gpus-per-node={self.gpus_per_node}"
    
    def run(self):
        global EXIT_FLAG
        workers = self.worker_group
        if self.total_nodes == 1:
            logger.warning("Running Serial job mode with only one node. Typically, "
            "there is only one rank per node, and Balsam master occupies the first node.\n"
            "Assuming you are testing; will run 4 ranks (1 master; 3 workers) on the node.\n"
            "This will cause heavy oversubscription with production workloads."
            )
            num_ranks = 4
            rpn = 4
        else:
            num_ranks = self.total_nodes
            rpn = 1
        mpi_str = workers.mpi_cmd(workers, app_cmd=self.app_cmd,
                               num_ranks=num_ranks, ranks_per_node=rpn,
                               cpu_affinity='none', envs={})
        logger.info(f'Starting MPI Fork ensemble process:\n{mpi_str}')

        self.outfile = open(os.path.join(settings.LOGGING_DIRECTORY, 'ensemble.out'), 'wb')
        self.process = subprocess.Popen(
            args=shlex.split(mpi_str),
            bufsize=1,
            stdout=self.outfile,
            stderr=subprocess.STDOUT,
            shell=False)

        while not EXIT_FLAG:
            try: 
                retcode = self.process.wait(timeout=2)
            except subprocess.TimeoutExpired: 
                pass
            else:
                logger.info(f'ensemble pull subprocess returned {retcode}')
                break

        self.process.terminate()
        try:
            self.process.wait(timeout=10)
        except subprocess.TimeoutExpired:
            self.process.kill()
        self.outfile.close()
        
def main(args):
    signal.signal(signal.SIGINT,  sig_handler)
    signal.signal(signal.SIGTERM, sig_handler)

    wf_filter = args.wf_filter
    job_mode = args.job_mode
    timelimit_min = args.time_limit_minutes
    nthread = (args.num_transition_threads if args.num_transition_threads
              else settings.NUM_TRANSITION_THREADS)
    gpus_per_node = args.gpus_per_node
    
    Launcher = MPILauncher if job_mode == 'mpi' else SerialLauncher
    
    try:
        transition_pool = transitions.TransitionProcessPool(nthread, wf_filter)
        launcher = Launcher(wf_filter, timelimit_min, gpus_per_node)
        launcher.run()
    except:
        raise
    finally:
        transition_pool.terminate()
        logger.info("Exit: Launcher exit graceful\n\n")

def get_args(inputcmd=None):
    '''Parse command line arguments'''
    parser = config_launcher_subparser()
    if inputcmd:
        return parser.parse_args(inputcmd)
    else:
        return parser.parse_args()

if __name__ == "__main__":
    setup()
    args = get_args()
    config_logging('launcher')
    logger.info("Loading Balsam Launcher")
    main(args)



import json
import os
import sys
import shutil
import tempfile
from balsam.django_config.serverinfo import ServerInfo

home_dir = os.path.expanduser('~')
BALSAM_HOME = os.path.join(home_dir, '.balsam')

def bootstrap():
    if not os.path.exists(BALSAM_HOME):
        os.makedirs(BALSAM_HOME, mode=0o755)

    sys.path.append(BALSAM_HOME)

    migrations_path = os.path.join(BALSAM_HOME, 'balsamdb_migrations')
    user_settings_path = os.path.join(BALSAM_HOME, 'settings.json')
    user_policy_path = os.path.join(BALSAM_HOME, 'theta_policy.ini')
    user_templates_path = os.path.join(BALSAM_HOME, 'job-templates')
    here = os.path.dirname(os.path.abspath(__file__))

    default_settings_path = os.path.join(here, 'default_settings.json')
    default_settings = json.load(open(default_settings_path))

    if not os.path.exists(user_settings_path):
        shutil.copy(default_settings_path, user_settings_path)
        print("Set up your Balsam config directory at", BALSAM_HOME)
    if not os.path.exists(user_policy_path):
        default_policy_path = os.path.join(here, 'theta_policy.ini')
        shutil.copy(default_policy_path, user_policy_path)
    if not os.path.exists(user_templates_path):
        default_templates_path = os.path.join(here, 'job-templates')
        shutil.copytree(default_templates_path, user_templates_path)
    if not os.path.exists(migrations_path):
        os.makedirs(migrations_path, mode=0o755)
        with open(os.path.join(migrations_path, '__init__.py'), 'w') as fp:
            fp.write('\n')
    try:
        user_settings = json.load(open(user_settings_path))
        for key in default_settings: assert key in user_settings
    except (AssertionError, json.decoder.JSONDecodeError):
        print(f"Detected invalid settings in {user_settings_path}; replacing with defaults!")
        shutil.copy(default_settings_path, user_settings_path)
        user_settings = json.load(open(user_settings_path))

    thismodule = sys.modules[__name__]
    for k, v in user_settings.items():
        setattr(thismodule, k, v)

def resolve_db_path():
    path = os.environ.get('BALSAM_DB_PATH', '')
    if path: path = os.path.abspath(os.path.expanduser(path))
    if not os.path.isdir(path):
        sys.stderr.write(f"Please use `source balsamactivate` to set BALSAM_DB_PATH to a valid location\n")
        sys.stderr.write(f"  --> `balsam which --list` recalls visited DBs in the filesystem\n")
        sys.stderr.write(f"Or, use `balsam init` to start a new Balsam DB\n")
        sys.exit(1)
    os.environ['BALSAM_DB_PATH'] = path
    return path

if os.environ.get('BALSAM_SPHINX_DOC_BUILD_ONLY', False):
    tempdir = tempfile.TemporaryDirectory()
    BALSAM_PATH = tempdir.name
    print("detected env BALSAM_SPHINX_DOC_BUILD_ONLY: will not connect to a real DB")
else:
    BALSAM_PATH = resolve_db_path()

bootstrap()
DATABASES = ServerInfo(BALSAM_PATH).django_db_config()

# SUBDIRECTORY SETUP
# ------------------
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
LOGGING_DIRECTORY = os.path.join(BALSAM_PATH , 'log') 
DATA_PATH = os.path.join(BALSAM_PATH ,'data')
SERVICE_PATH = os.path.join(BALSAM_PATH ,'qsubmit')
BALSAM_WORK_DIRECTORY = DATA_PATH

for d in [
      BALSAM_PATH ,
      DATA_PATH,
      LOGGING_DIRECTORY,
      SERVICE_PATH
]:
    if not os.path.exists(d):
        os.makedirs(d)

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '=gyp#o9ac0@w3&-^@a)j&f#_n-o=k%z2=g5u@z5+klmh_*hebj'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = [
    'balsam.core.apps.BalsamCoreConfig',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
]

MIGRATION_MODULES = {'core': 'balsamdb_migrations'}

MIDDLEWARE_CLASSES = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'balsam.django_config.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'balsam.django_config.wsgi.application'


# Password validation
# https://docs.djangoproject.com/en/1.9/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')



#!/Users/misha/anaconda3/envs/testmpi/bin/python
import balsam.launcher.dag as dag

print("hello from square_post")
print(f"jobid: {dag.current_job.pk}")
if dag.ERROR:
    print("recognized error")
    dag.current_job.update_state("JOB_FINISHED", "handled error in square_post")
if dag.TIMEOUT:
    print("recognized timeout")
    dag.current_job.update_state("JOB_FINISHED", "handled timeout in square_post")



import json
import os

def refresh_db_index():
    home_dir = os.path.expanduser('~')
    BALSAM_HOME = os.path.join(home_dir, '.balsam')

    index_path = os.path.join(BALSAM_HOME, 'databases.json')

    if os.path.exists(index_path):
        with open(index_path) as fp: db_list = json.load(fp)
    else:
        db_list = []
        if not os.path.exists(BALSAM_HOME):
            os.makedirs(BALSAM_HOME)

    cur_db = os.environ.get('BALSAM_DB_PATH')
    if cur_db:
        cur_db = os.path.abspath(os.path.expanduser(cur_db))
        if cur_db not in db_list: db_list.append(cur_db)

    for i, db in reversed(list(enumerate(db_list[:]))):
        if not os.path.exists(db): del db_list[i]
    with open(index_path, 'w') as fp: json.dump(db_list, fp, indent=1)
    return db_list



import os
import sys
from importlib.util import find_spec

from balsam.service.models import BalsamJob
from tests.BalsamTestCase import BalsamTestCase
from tests.BalsamTestCase import create_job, create_app

from tests import util
    
class TestInsertion(BalsamTestCase):
    def setUp(self):
        from itertools import takewhile, product

        hello = find_spec("tests.benchmarks.concurrent_insert.hello").origin
        create_app(name="hello", executable=hello)
        self.launcherInfo = util.launcher_info()
        
        max_workers = self.launcherInfo.num_workers
        worker_counts = list(takewhile(lambda x: x<=max_workers, (2**i for i in range(20))))
        if max_workers not in worker_counts:
            worker_counts.append(max_workers)
        worker_counts = list(reversed(worker_counts))
        #ranks_per_node = [4, 8, 16, 32]
        ranks_per_node = [32]
        self.experiments = product(worker_counts, ranks_per_node)
        
        # Load mpi4py/Balsam on compute nodes prior to experiments
        hello = find_spec("tests.benchmarks.concurrent_insert.hello").origin
        python = sys.executable
        app_cmd = f"{python} {hello}"
        mpi_str = self.launcherInfo.mpi_cmd(
            self.launcherInfo.workerGroup.workers,
            app_cmd=app_cmd,
            envs={},
            num_ranks=max_workers,
            ranks_per_node=1,
            threads_per_rank=1,
            threads_per_core=1
        )
        stdout, elapsed_time = util.cmdline(mpi_str)

    def test_concurrent_mpi_insert(self):
        '''Timing: many MPI ranks simultaneously call dag.add_job'''
        resultpath = util.benchmark_outfile_path('concurrent_insert.dat')

        title = 'test_concurrent_mpi_insert'
        comment = 'Each rank simultaneously calls dag.add_job (num_ranks simultaneous insertions)'
        resultTable = util.FormatTable(
            'num_nodes ranks_per_node num_ranks total_time_sec'.split()
        )

        python = sys.executable
        insert_app = find_spec("tests.benchmarks.concurrent_insert.mpi_insert").origin

        for (num_nodes, rpn) in self.experiments:
            BalsamJob.objects.all().delete()

            total_ranks = num_nodes * rpn

            app_cmd = f"{python} {insert_app}"
            mpi_str = self.launcherInfo.mpi_cmd(
                self.launcherInfo.workerGroup.workers,
                app_cmd=app_cmd,
                envs={},
                num_ranks=total_ranks,
                ranks_per_node=rpn,
                threads_per_rank=1,
                threads_per_core=1
            )
            stdout, elapsed_time = util.cmdline(mpi_str)
            
            success = list(l for l in stdout.split('\n') if 'added job: success' in l)
            self.assertEqual(len(success), total_ranks)
            self.assertEqual(BalsamJob.objects.count(), total_ranks)

            resultTable.add_row(num_nodes=num_nodes, ranks_per_node=rpn,
                                num_ranks=total_ranks,
                                total_time_sec=elapsed_time
                                )
        
            with open(resultpath, 'w') as fp:
                fp.write(resultTable.generate(title, comment))



from setuptools import setup, Extension
from setuptools.config import read_configuration
from os import path, environ

here = path.dirname(__file__)
conf_dict = read_configuration(path.join(here, 'setup.cfg'))

install_requires = conf_dict['options']['install_requires']

if environ.get('READTHEDOCS') == 'True':
    install_requires = [k for k in install_requires if 'mpi4py' not in k]

#extensions = [
#    Extension(
#        "balsam.service.pack._packer",
#        ["balsam/service/pack/_packer.pyx"]
#    ),
#]

setup(
    install_requires = install_requires,
    #ext_modules = extensions,
)



from balsam.core import models
from os import environ
from os.path import basename
import uuid
from django.core.exceptions import FieldError
Job = models.BalsamJob
AppDef = models.ApplicationDefinition
QueuedLaunch = models.QueuedLaunch

try: LIMIT = int(environ.get('BALSAM_LS_LIMIT', 5000))
except ValueError: LIMIT = 5000

def app_string(app):
    return " ".join(basename(p) for p in app.split())

def print_table(queryset, fields, field_header={}, transforms={}):
    fields = list(fields)
    num_fields = len(fields)
    records = list(queryset.values_list(*fields))

    # rename columns in field_header dict to the desired values
    for field_name, header_name in field_header.items():
        i = fields.index(field_name)
        fields[i] = header_name

    # apply transforms on each column
    transforms = {fields.index(field_name) : fxn for field_name,fxn in transforms.items()}
    for irow, row in enumerate(records):
        row = list(map(str, row))
        for i, fxn in transforms.items():
            row[i] = fxn(row[i])
        records[irow] = row

    widths = [max((len(row[field]) for row in records)) for field in range(num_fields)]
    widths = [max(w,len(f)) for w,f in zip(widths, fields)] 
    format = ' | '.join(f'%{width}s' for width in widths)
    header = format % tuple(fields)
    print(header)
    print('-'*len(header))
    for row in records:
        print(format % tuple(f.ljust(w) for f,w in zip(row, widths)))

def print_history(jobs):
    count = jobs.count()
    query = jobs[:LIMIT].values_list('name', 'job_id', 'state_history')
    for (name,job_id,state_history) in query:
        print(f'Job {name} [{job_id}]')
        print(f'------------------------------------------------')
        print(f'{state_history}\n')
    if count > LIMIT: print(f"({count-LIMIT} more BalsamJobs not shown...)")

def print_jobs(jobs, verbose):
    count = jobs.count()
    jobs = jobs[:LIMIT]
    if verbose:
        for job in jobs: print(job)
        if count > LIMIT: print(f"({count-LIMIT} more BalsamJobs not shown...)")
        return

    fields = ['job_id', 'name', 'workflow', 'application', 'state']
    add = environ.get('BALSAM_LS_FIELDS')
    if add: fields.extend([s.strip() for s in add.split(':') if s])
    transforms = {'application' : app_string}
    try:
        print_table(jobs, fields, transforms=transforms)
        if count > LIMIT: print(f"({count-LIMIT} more BalsamJobs not shown...)")
    except FieldError as e:
        print("***  You specified a nonexistant field in BALSAM_LS_FIELDS; please unset this ***")
        print(e)
        return

def print_by_states(jobs):
    from django.db.models import Count
    states = jobs.values('state').annotate(s_count=Count('state'))
    for result in states:
        state, count = result['state'], result['s_count']
        print(f"{state.rjust(16)}  {count}")

def print_subtree(job, indent=1):
    def job_str(job): return f"{job.name:10} {job.cute_id}"
    print('|'*indent, end=' ')
    print(5*indent*' ', job_str(job))
    for job in job.get_children():
        print_subtree(job, indent+1)

def print_jobs_tree(jobs):
    count = jobs.count()
    roots = jobs.filter(parents='[]')[:LIMIT]
    for job in roots: print_subtree(job)
    if count > LIMIT:
        print(f"(Omitted {count-LIMIT} jobs from display...)")

def ls_jobs(namestr, show_history, jobid, verbose, tree, wf, state, by_states):
    results = Job.objects.all()
    if namestr: results = results.filter(name__icontains=namestr)

    if jobid and len(jobid) < 36:
        results = results.filter(job_id__icontains=jobid)
    elif jobid:
        try:
            pk = uuid.UUID(jobid.strip())
            results = results.filter(job_id=pk)
        except ValueError:
            results = results.filter(job_id__icontains=jobid)

    if wf: results = results.filter(workflow__icontains=wf)
    if state: results = results.filter(state=state)
    
    if not results.exists():
        print("No jobs found matching query")
        return

    if show_history: print_history(results)
    elif tree: print_jobs_tree(results)
    elif by_states: print_by_states(results)
    else: print_jobs(results, verbose)

def ls_queues(verbose):
    from balsam.service.schedulers.exceptions import StatusNonZeroReturnCode
    try:
        QueuedLaunch.refresh_from_scheduler()
    except StatusNonZeroReturnCode as e:
        print(f"Scheduler is not reachable: {e}")
        return
    allq = QueuedLaunch.objects.all()
    if verbose:
        for q in allq: print(q)
        return
    if not allq.exists():
        print("No queued jobs detected")
        return

    fields = ['scheduler_id', 'project', 'queue', 'nodes', 'wall_minutes', 'state', 'job_mode', 'command', 'from_balsam']
    header = {'command' : 'Command', 'from_balsam' : 'Submitted from Balsam?'}
    transforms = {'Command' : lambda x: basename(x)}
    print_table(allq, fields, header, transforms)

def ls_apps(namestr, appid, verbose):
    if namestr:
        results = AppDef.objects.filter(name__icontains=namestr)
    elif appid:
        results = AppDef.objects.filter(job_id__icontains=appid)
    else:
        results = AppDef.objects.all()
    if not results:
        print("No apps found matching query")
        return
    if verbose:
        for app in results: print(app)
        return

    fields = ('pk', 'name', 'executable', 'description')
    transforms = {'executable' : app_string}
    print_table(results, fields, transforms=transforms)


def ls_wf(name, verbose, tree, wf):
    workflows = Job.objects.order_by().values('workflow').distinct()
    workflows = [wf['workflow'] for wf in workflows]

    if wf: name = wf # wf argument overrides name

    if name and name not in workflows:
        print(f"No workflow matching {name}")
        return

    if name and name in workflows: 
        workflows = [name]
        verbose = True

    print("Workflows")
    print("---------")
    for wf in workflows:
        print(wf)
        if tree:
            print('-'*len(wf))
            jobs_by_wf = Job.objects.filter(workflow=wf)
            print_jobs_tree(jobs_by_wf)
            print()
        elif verbose:
            print('-'*len(wf))
            jobs_by_wf = Job.objects.filter(workflow=wf)
            print_jobs(jobs_by_wf, False)
            print()



import os
import sys
import random
import time

class InfoLock:
    EXPIRATION_SECONDS = 20

    def __init__(self, path):
        assert os.path.isdir(path)
        self.lock_path = os.path.join(path, '.infolock')

    def check_stale(self):
        try: mtime = os.path.getmtime(self.lock_path)
        except OSError: return False
        else: return (time.time() - mtime > self.EXPIRATION_SECONDS)

    def acquire_lock(self, timeout=40.0):
        acquired = False
        start = time.time()
        newline = False
        def _time_left():
            if timeout is None: return True
            return time.time() - start < timeout

        while not acquired and _time_left():
            try: os.mkdir(self.lock_path)
            except OSError:
                time.sleep(1.0 + random.uniform(0, 0.5))
                print(".", end='', flush=True)
                newline = True
                if self.check_stale():
                    try: os.rmdir(self.lock_path)
                    except FileNotFoundError: pass
            else: 
                acquired = True
                if newline: print("")
        if not acquired:
            raise TimeoutError(f'Failed to acquire {self.lock_path} for {timeout} sec')

    def release_lock(self):
        try: os.rmdir(self.lock_path)
        except FileNotFoundError: pass
            
    def __enter__(self):
        self.acquire_lock()
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.release_lock()



import os
import sys
import time
from socket import gethostname
from balsam.service.schedulers.exceptions import *

import logging
logger = logging.getLogger(__name__)

class JobEnvironment:
    RECOGNIZED_HOSTS = {
        'BGQ'    : 'vesta cetus mira'.split(),
        'THETA'   : 'theta'.split(),
        'COOLEY' : 'cooley cc'.split()
    }

    def __init__(self, scheduler):
        self.scheduler_vars = scheduler.SCHEDULER_VARIABLES
        self.pid = os.getpid()
        self.hostname = gethostname()
        self.host_type = 'DEFAULT'
        
        self.current_scheduler_id = None
        self.num_workers = 1
        self.workers_str = None
        self.workers_file = None
        self.remaining_seconds = float("inf")
        self.get_env()
        try:
            info = scheduler.get_status(self.current_scheduler_id)
            self.remaining_seconds = info['time_remaining_sec']
        except (NoQStatInformation, TypeError, KeyError):
            pass
        self._last_check_seconds = time.time()

        for host_type, known_names in self.RECOGNIZED_HOSTS.items():
            if any(self.hostname.find(name) >= 0 for name in known_names):
                self.host_type = host_type
        logger.debug(f"Recognized host_type: {self.host_type}")

    def get_env(self):
        '''Check for environment variables (e.g. COBALT_JOBID) indicating 
        currently inside a scheduled job'''
        for generic_name, specific_var in self.scheduler_vars.items():
            value = os.environ.get(specific_var, None)
            if value is not None and value.startswith('num'): value = int(value)
            setattr(self, generic_name, value)

        if self.current_scheduler_id:
            self.current_scheduler_id = int(self.current_scheduler_id)
            logger.debug(f"Detected scheduler ID {self.current_scheduler_id}")

    def remaining_time_seconds(self):
        '''Either counts down from RemainingTime obtained from scheduler, or infinity'''
        now = time.time()
        elapsed_time = now - self._last_check_seconds
        self.remaining_seconds -= elapsed_time
        self._last_check_seconds = now
        return self.remaining_seconds
